/*****************************************************************************
  Pocket BASIC

  |PRBASICC^[v^

  Choi Kyun

  2009-10

  ver 0.1.0
*****************************************************************************/



#include <PalmOS.h>
#include <PalmOSGlue.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

/* MathLib headers */
#include "MathLib.h"

#include "tiny_basic.h"
#include "tiny_basic_Rsc.h"
#include "basic.h"



/*********************************************************************
 * Entry Points
 *********************************************************************/
void disp_error( int );
char get_char( char ** );
int get_line_number( char **);
Boolean conv_il( char **, char *, int );
int find_edit_line( int );
Boolean insert_pa( int, char *, int );
void conv_basic_to_ascii( int );

char get_token_char( char ** );
TokenType get_token( char ** );
double expression( char ** );
double term( char ** );
double factor( char ** );
void back_token( char ** );
int get_array_index( char ** );
void cmd_let( char ** );
void cmd_let2( int, char ** );
void cmd_let_str( int, char ** );
void check_end( char * );
void disp_edit_line( void );
Boolean logic( char ** );
int copy_str( char **, char * );
Boolean str_logic( char ** );
int str_expression( char **, char * );
int str_factor( char **, char * );
void palm_put_str( char * );
void palm_put_num( double );
int find_goto_line( int );
double round_n( double, int );
void input_end( void );

void cmd_let_str_d( char ** );
void cmd_end( char ** );
void cmd_vac( char ** );
void cmd_locate( char ** );
void cmd_if( char ** );
void cmd_print( char ** );
void cmd_input( char ** );
void cmd_cls( char ** );
void cmd_goto( char ** );
void cmd_gosub( char ** );
void cmd_return( char ** );
void cmd_for( char ** );
void cmd_to( char ** );
void cmd_next( char ** );
void cmd_color( char ** );
UInt8 cmd_key( void );

int cmd_len( char ** );
int cmd_mid( char **, char * );
double cmd_val( char ** );
int cmd_str( char **, char * );



/*********************************************************************
 * Internal Constants
 *********************************************************************/
KeywordType keyword[] = {
  { "CSR", TOKEN_LOCATE },
  { "PRINT", TOKEN_PRINT },
  { "INPUT", TOKEN_INPUT },
  { "IF", TOKEN_IF },
  { "THEN", TOKEN_THEN },
  { "GOTO", TOKEN_GOTO },
  { "GOSUB", TOKEN_GOSUB },
  { "RETURN", TOKEN_RETURN },
  { "FOR", TOKEN_FOR },
  { "TO", TOKEN_TO },
  { "NEXT", TOKEN_NEXT },
  { "STEP", TOKEN_STEP },
  { "VAC", TOKEN_VAC },
  { "BEEP", TOKEN_BEEP },
  { "END", TOKEN_END },
  { "LET", TOKEN_LET },
  { "'", TOKEN_REM },
  { "COLOR", TOKEN_COLOR },
  { "CLS", TOKEN_CLS },

  { "LEN", TOKEN_LEN },
  { "MID", TOKEN_MID },
  { "VAL", TOKEN_VAL },
  { "STR$", TOKEN_STR },

  { "KEY", TOKEN_KEY },
  { "RAN#", TOKEN_RAN },
  { "SIN", TOKEN_SIN },
  { "COS", TOKEN_COS },
  { "TAN", TOKEN_TAN },
  { "ASN", TOKEN_ASN },
  { "ACS", TOKEN_ACS },
  { "ATN", TOKEN_ATN },
  { "LOG", TOKEN_LOG },
  { "LN", TOKEN_LN },
  { "EXP", TOKEN_EXP },
  { "SQR", TOKEN_SQR },
  { "ABS", TOKEN_ABS },
  { "SGN", TOKEN_SGN },
  { "INT", TOKEN_INT },
  { "FRAC", TOKEN_FRAC },
  { "TICK", TOKEN_TICK },

  { "\"", TOKEN_STRING },
  { ",", TOKEN_COMMA },
  { ";", TOKEN_SEMICOLON },
  { ":", TOKEN_COLON },
  { "+", TOKEN_PLUS },
  { "-", TOKEN_MINUS },
  { "*", TOKEN_ASTER },
  { "/", TOKEN_SLASH },
  { "%", TOKEN_PERCENT },
  { "(", TOKEN_PAREN_B },
  { ")", TOKEN_PAREN_E },

  { "=", TOKEN_EQ },
  { "\xb5", TOKEN_NEQ },
  { "\xb6", TOKEN_LTE },
  { "<", TOKEN_LT },
  { "\xb7", TOKEN_GTE },
  { ">", TOKEN_GT },
  { "\xb0", TOKEN_PI },

  { "", 0 }
};



/*********************************************************************
 * Internal Functions
 *********************************************************************/

// Xe[ggs
// read_p Xe[gg
void
execute( void )
{
  TokenType t;
  int align;

//  if( TimGetTicks() < g_last_tick + g_frame_tick ) return;
  // TickXV
//  g_last_tick = TimGetTicks();

  for(;;)
    {
      if( g_program_pointer == NULL ) return; // I

      get_token_char( &g_program_pointer );
      if( *g_program_pointer == ':' )
        {
          g_program_pointer++;
          continue;
        }

      if( (UInt8) *(g_program_pointer) == TOKEN_REM )
        {
          while( *g_program_pointer ) g_program_pointer++;
        }

      if( *g_program_pointer == '\0' )
        {
          if( (UInt32)g_program_pointer & 1 ) // ACg
            align = 1;
          else
            align =  2;

          if( g_program_pointer >= &g_pa_p[ g_current_pa ][0]
              && g_program_pointer + 4 + align < &g_pa_p[ g_current_pa ][0] + g_pa_last_ptr[ g_current_pa ] )
            {
              g_line_top = g_program_pointer + align;
              g_program_pointer = g_line_top + sizeof(int) + sizeof(char);
              continue;
            }
        }
      break;
    }

  // R}hs
  t = get_token( &g_program_pointer );
  if( g_err_number ) return;
  switch( t.code )
    {
    case TOKEN_NULL: cmd_end( &g_program_pointer ); g_program_pointer = NULL; break;
    case TOKEN_VAR: // ϐ
    case TOKEN_VAR_S:
    case TOKEN_VAR_D: back_token( &g_program_pointer ); cmd_let( &g_program_pointer ); break;
    case TOKEN_LET: cmd_let( &g_program_pointer ); break;
    case TOKEN_IF: cmd_if( &g_program_pointer ); break;
    case TOKEN_GOTO: cmd_goto( &g_program_pointer ); break;
    case TOKEN_GOSUB: cmd_gosub( &g_program_pointer ); break;
    case TOKEN_RETURN: cmd_return( &g_program_pointer ); break;
    case TOKEN_PRINT: cmd_print( &g_program_pointer ); break;
    case TOKEN_INPUT: cmd_input( &g_program_pointer ); break;
    case TOKEN_FOR: cmd_for( &g_program_pointer ); break;
    case TOKEN_TO: cmd_to( &g_program_pointer ); break;
    case TOKEN_NEXT: cmd_next( &g_program_pointer ); break;
    case TOKEN_LOCATE: cmd_locate( &g_program_pointer ); break;
    case TOKEN_COLOR: cmd_color( &g_program_pointer ); break;
    case TOKEN_END: cmd_end( &g_program_pointer ); break; // I
    case TOKEN_VAC: cmd_vac( &g_program_pointer ); break;
    case TOKEN_CLS: cmd_cls( &g_program_pointer ); break;
    case TOKEN_ERROR: break;
    default:
      disp_error( ERR_SE );
      return;
    }
  return;
}



// ̎擾
// I |C^ ́A擾wB
static char
get_char( char **read_p )
{
  char c;

  for(;;)
    {
      c = **read_p;
      if( (c <= '\0' || c > ' ') ) break; // 󔒃XLbv
      (*read_p)++;
    }

  c = toupper( c );
  return c;
}



// g[Np1擾
// I |C^ ́A擾wB
static char
get_token_char( char **read_p )
{
  char c;


  c = **read_p;
  if( c==':' || c=='\'' ) c = 0;

  return c;
}



// g[N擾
// Ԍ  l  ϐ  ̑LN^
// |C^͎̃g[Nw
static TokenType
get_token( char **read_p )
{
  TokenType t = { 0, 0 };
  char c;
  double power;


  c = get_token_char( read_p );
  g_token_top = *read_p;

  if( c == '\0' )
    {
      t.code = TOKEN_NULL;
    }
  else if( c >= '0' && c <= '9' ) // l double
    {
      t.code = TOKEN_NUMBER;
      t.value = power = 0.0;
      for(;;)
        {
          t.value = t.value * 10.0 + (c - '0');
          power *= 10.0;
          if( power > 10000000000.0 || t.value > DBL_MAX ) // double 10܂ŔF
            {
              t.code = TOKEN_ERROR;
              disp_error( ERR_DO );
              return t;
            }
          (*read_p)++;
          c = get_token_char( read_p );
          if( c == '.' )
            {
              power = 1.0;
              (*read_p)++;
              c = get_token_char( read_p );
            }
          if( c < '0' || c > '9' ) break;
        }
      if( power )
        t.value /= power;
    }
  else if( c >= 'A' && c <= 'Z' ) // ϐ
    {
      (*read_p)++;
      if( get_token_char( read_p ) == '$' )
        {
          (*read_p)++;
          t.code = TOKEN_VAR_S;
          t.value = (int) c - 'A';
        }
      else
        {
          t.code = TOKEN_VAR;
          t.value = (int) c - 'A';
        }
    }
  else if( c == '$' )
    {
      (*read_p)++;
      t.code = TOKEN_VAR_D; // $ϐ
      t.value = 0;
    }
  else
    {
      (*read_p)++;
      t.code = (UInt8)c; // Ԍ  LN^
    }

  return t;
}



// _
static Boolean
logic( char **read_p )
{
  double  ans, rans;
  Boolean b;
  TokenType t;

  ans = expression( read_p );
  if( g_err_number ) return 0;

  ans = round_n( ans, ROUND_POW ); // ۂ߂

  t = get_token( read_p );
  if( g_err_number ) return 0;

  if( t.code < TOKEN_ROPMIN || t.code > TOKEN_ROPMAX )
    {
      back_token( read_p );
      disp_error( ERR_SE );
      return 0;
    }

  rans = expression( read_p );

  if( g_err_number ) return 0;

  rans = round_n( rans, ROUND_POW );

  switch( t.code )
    {
    case TOKEN_EQ:  b = ans == rans; break;
    case TOKEN_NEQ: b = ans != rans; break;
    case TOKEN_LT:  b = ans <  rans; break;
    case TOKEN_LTE: b = ans <= rans; break;
    case TOKEN_GT:  b = ans >  rans; break;
    case TOKEN_GTE: b = ans >= rans; break;
    }
  return b;
}



// 
static double
expression( char **read_p )
{
  double ans, rans;


  ans = term( read_p );
  if( g_err_number ) return 0;

  for(;;)
    {
      TokenType t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PLUS && t.code != TOKEN_MINUS )
        {
          back_token( read_p );
          break;
        }
      rans = term( read_p );
      if( g_err_number ) return 0;

      switch( t.code )
        {
        case TOKEN_PLUS: ans += rans; break;
        case TOKEN_MINUS: ans -= rans; break;
        }
    }
  return ans;
}



// 
static double
term( char **read_p )
{
  double ans, rans;


  ans = factor( read_p );
  if( g_err_number ) return 0;

  for(;;){
    TokenType t = get_token( read_p );
    if( g_err_number ) return 0;
    if( t.code != TOKEN_ASTER && t.code != TOKEN_SLASH && t.code != TOKEN_PERCENT )
      {
        back_token( read_p );
        break;
      }

    rans = factor( read_p );
    if( g_err_number ) return 0;

    switch( t.code )
      {
      case TOKEN_SLASH:
      case TOKEN_PERCENT:
        if( rans == 0 )
          {
            disp_error( ERR_DZ );
            return 0;
          }
        break;
      }

    switch( t.code )
      {
      case TOKEN_ASTER: ans *= rans; break;
      case TOKEN_SLASH: ans /= rans; break;
      case TOKEN_PERCENT: ans = fmod( ans, rans ); break;
      }
  }

  return ans;
}



// q
static double
factor( char **read_p )
{
  double v;
  Int32 i;
  TokenType t = get_token( read_p );

  if( g_err_number ) return 0;

  switch( t.code )
    {
    case TOKEN_MINUS: return -factor( read_p );
    case TOKEN_NUMBER: return t.value;

    case TOKEN_VAR:
      v = get_array_index( read_p );
      if( g_err_number ) return 0;

        if( (int)(t.value + v) < 0 || (int)(t.value + v) > 26 )
        {
          disp_error( ERR_OR );
          return 0;
        }
      if( g_variable[ (int)(t.value + v) ].type == VARIABLE_STR ) // lϐ̂
        {
          disp_error( ERR_TM );
          return 0;
          }
      return g_variable[ (int)(t.value + v) ].contents.val; // ϐ

      // l֐
    case TOKEN_SIN:
      v = sin( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_COS:
      v = cos( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_TAN:
      v = tan( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_ASN:
      v = asin( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_ACS:
      v = acos( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_ATN:
      v = atan( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_LOG:
      v = log10( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_LN:
      v = log( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_EXP:
      v = exp( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_SQR:
      v = sqrt( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_ABS:
      v = factor( read_p );
      if( g_err_number ) return 0;
      return Abs( v );
    case TOKEN_SGN:
      v =  factor( read_p );
      if( g_err_number ) return 0;
      if( v > 0 )
        return 1;
      else if( v < 0 )
        return -1;
      break;
    case TOKEN_INT:
      v = floor( factor( read_p ) );
      if( g_err_number ) return 0;
      return v;
    case TOKEN_FRAC:
      v = factor( read_p );
      if( g_err_number ) return 0;
      return v - (int) v;
    case TOKEN_TICK:
      return TimGetTicks();
    case TOKEN_RAN:
      i = (int) factor( read_p );
      if( i > sysRandomMax )
        i = sysRandomMax;
      return RND( i );
    case TOKEN_PI:
      return PI;

      // ֐
    case TOKEN_LEN:
      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_B )// ʕKv
        {
          disp_error( ERR_SE );
          return 0;
        }

      v = cmd_len( read_p );

      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_E )// 
        {
          disp_error( ERR_SE );
          return 0;
        }
      return v;

    case TOKEN_VAL:
      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_B )// ʕKv
        {
          disp_error( ERR_SE );
          return 0;
        }

      v = cmd_val( read_p );

      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_E )// 
        {
          disp_error( ERR_SE );
          return 0;
        }
      return v;

      // 
    case TOKEN_PAREN_B:
      v = expression( read_p );
      if( g_err_number ) return 0;

      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_E )
        {
          //disp_error( ERR_SE );
          //return 0;
          back_token( read_p );// ʏȗ
        }
      return v;

    default:
      disp_error( ERR_SE );
      return 0;
    }

  return 0;
}



// r
static Boolean
str_logic( char **read_p )
{
  Boolean b;
  int ans_len, rans_len;
  TokenType t;
  char ans[ MAX_STRING_VAR+1 ];
  char rans[ MAX_STRING_VAR+1 ];

  ans[0] = '\0';rans[0] = '\0';
  ans_len = str_expression( read_p, ans );
  if( g_err_number ) return 0;

  t = get_token( read_p );
  if( g_err_number ) return 0;

  if( t.code != TOKEN_EQ && t.code != TOKEN_NEQ )
    {
      back_token( read_p );
      disp_error( ERR_SE );
      return 0;
    }

  rans_len = str_expression( read_p, rans );
  if( g_err_number ) return 0;

  switch( t.code )
    {
    case TOKEN_EQ:
      if( ! StrCompareAscii( ans, rans ) )
        b = true;
      else
        b = false;
      break;
    case TOKEN_NEQ:
      if( StrCompareAscii( ans, rans ) )
        b = true;
      else
        b = false;
      break;
    }
  return b;
}



// 񉉎Z
static int
str_expression( char **read_p, char *s )
{
  int ans_len, rans_len;
  char ans[ MAX_STRING_VAR+1 ];
  char rans[ MAX_STRING_VAR+1 ];


  s[0] = '\0'; ans[0] = '\0';
  ans_len = str_factor( read_p, ans );
  StrNCat( s, ans, MAX_STRING_VAR+1 );

  if( g_err_number ) return 0;

  for(;;)
    {
      TokenType t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PLUS )
        {
          back_token( read_p );
          break;
        }
      rans[0] = '\0';
      rans_len = str_factor( read_p, rans );
      if( g_err_number ) return 0;

      switch( t.code )
        {
        case TOKEN_PLUS:
          StrNCat( s, rans, MAX_STRING_VAR+1 );
          ans_len += rans_len;
          if( ans_len >= MAX_STRING_VAR )
            {
              ans_len = MAX_STRING_VAR;
              disp_error( ERR_LG );
            }
          break;
        }
    }
  return ans_len;// Ԃ
}



// q  "", a$-z$, a$()-z$(), $
static int
str_factor( char **read_p, char *s )
{
  int v, len;
  TokenType t = get_token( read_p );

  if( g_err_number ) return 0;
  s[0] = '\0';
  switch( t.code )
    {
    case TOKEN_VAR_S:
      v = get_array_index( read_p );
      if( g_err_number ) return 0;

      if( (int)(t.value + v) < 0 || (int)(t.value + v) > 26 )
        {
          disp_error( ERR_OR );
          return 0;
        }
      if( g_variable[ (int)(t.value + v) ].type == VARIABLE_NUM ) // ϐ̂
        {
          disp_error( ERR_TM );
          return 0;
        }
      StrCat( s, &g_variable[ (int)(t.value + v) ].contents.str[0] ); // ӂ̕ϐ
      len = StrLen( s );
      break;

    case TOKEN_VAR_D:
      StrCat( s, g_special_strings ); // ϐp
      len = StrLen( s );
      break;

    case TOKEN_STRING:
      len = copy_str( read_p, s ); // " "
      break;

    case TOKEN_KEY: //L[R[h
      len = 1;
      s[0] = cmd_key();
      s[1] = '\0';
      break;

      // ֐
    case TOKEN_MID:
      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_B ) // ʕKv
        {
          disp_error( ERR_SE );
          return 0;
        }

      len = cmd_mid( read_p, s );

      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_E ) // 
        {
          disp_error( ERR_SE );
          return 0;
        }
      break;

    case TOKEN_STR:
      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_B ) // ʕKv
        {
          disp_error( ERR_SE );
          return 0;
        }

      len = cmd_str( read_p, s );

      t = get_token( read_p );
      if( g_err_number ) return 0;
      if( t.code != TOKEN_PAREN_E ) // 
        {
          disp_error( ERR_SE );
          return 0;
        }
      break;

    default:
      disp_error( ERR_SE );
      return 0;
    }

  return len;
}



static void
back_token( char **read_p )
{
  *read_p = g_token_top;
}



// z̓Y
// PB-100 A(0) = A, A(1) = B, A(2) = C... Ƃ
static int
get_array_index( char **read_p )
{
  int v;
  TokenType t = get_token( read_p );

  if( g_err_number ) return 0;
  if( t.code != TOKEN_PAREN_B )
    {
      back_token( read_p ); // z񂶂Ȃ
      return 0;
    }

  v = expression( read_p );
  if( g_err_number ) return 0;

  t = get_token( read_p );
  if( g_err_number ) return 0;
  if( t.code != TOKEN_PAREN_E )
    {
      disp_error( ERR_SE );
      return 0;
    }

  return v;
}



// ۂ http://www.motorwarp.com/koizumi/round_c.html
static double
round_n( double src, int n )
{
  double dst;

  dst = src * pow(10, -n - 1);
  dst = rint( dst );

  return dst * pow(10, n + 1);
}



// ϐ̈揉
void
init_variable( void )
{
  int i;


  // ϐ
  for( i = 0; i < 26; i++ )
    {
      g_variable[i].type = NOT_USE;
      g_variable[i].contents.val = 0.0;
    }

  // ϐ
  g_special_strings[0] = '\0';
}



// sԍ擾
static int
get_line_number( char **read_p )
{
  char c = NULL;
  int val = 0;


  c = get_char( read_p );

  if( c =='\0' ) return val = -1;

  if( c >= '0' && c <= '9' )
    {
      for(;;)
        {
          val = val * 10 + (c - '0');
          if( val > MAX_LINE_NUMBER ) val = 0;
          (*read_p)++;
          c = get_char( read_p  );
          if( c < '0' || c > '9' ) break;
        }
    }
  return val;
}



// s܂ŒԌɕϊ
static Boolean
conv_il( char **read_p, char *new_line_buf, int new_line_size )
{
  char c;
  char *top;
  int i, j;
  GetCharModeType mode = chr_nomal;


  c = get_char( read_p );
  if( c =='\0' ) return false; // sԍ̂݃G[
  top = *read_p;

  do
    {
      for( i = 0; keyword[ i ].word[ 0 ] != 0; i++ )      // basicR}h
        {
          *read_p = top;
          for( j = 0; j < 7; j++ )
            {
              if( keyword[ i ].word[ j ] == 0 )
                {
                  c = keyword[i].code; // }b`
                  goto next;
                }
              if( get_char( read_p ) != keyword[ i ].word[ j ] ) break;
              (*read_p)++;
            }
        }
      // basicR}hȂꍇ͂̂܂܏o
      *read_p = top;
      c = get_char( read_p );
      (*read_p)++;
next:
      *(new_line_buf + new_line_size ) = c;
      new_line_size++;
      if( (UInt8)c == TOKEN_REM )
        {
          for(;;)
            {
              if( **read_p == '\0' ) break; // s܂ŃRs[
              *(new_line_buf + new_line_size++) = *(*read_p)++;
            }
          break;
        }
      else if( c == TOKEN_STRING )
        {
          for(;;)
            {
              if( **read_p == '\"' || **read_p == '\0' ) break; // " ܂ŃRs[
              *(new_line_buf + new_line_size++) = *(*read_p)++;
            }
          *(new_line_buf + new_line_size++) = *(*read_p)++;
        }
      top = *read_p;
    }
  while( get_char( read_p ) != '\0' );


  *(new_line_buf + new_line_size ) = '\0'; // sI}[Nǉ \0
  new_line_size++;
  // ACg낦
  if( new_line_size & 1 )
    {
      *(new_line_buf + new_line_size ) = '\0';
      new_line_size++;
    }
  new_line_buf[ 2 ] = (char) new_line_size;

  return true;
}



// vOGAɑ}
static Boolean
insert_pa( int line_number, char *new_line_buf, int new_line_size )
{
  int index, src, dst;
  Boolean ins = false;


  if( line_number == g_pa_last_number[ g_current_pa ] ) // ŏIs̏㏑
    ins = true;

  index = find_edit_line( line_number ); // sԍ

  delete_line( line_number ); // s폜

  if( ins ) g_pa_last_number[ g_current_pa ] = line_number;

  //󂫂
  src = g_pa_last_ptr[ g_current_pa ]; // vOGAŌ
  dst = g_pa_last_ptr[ g_current_pa ] + new_line_size;

  if( dst + sizeof(int) > MAX_PROGRAM_SIZE )
    {
      return false; // s
    }

  g_pa_last_ptr[ g_current_pa ] = dst; // TCYXV

  while( src > index )
    {
      g_pa_p[ g_current_pa ][ --dst ] = g_pa_p[ g_current_pa ][ --src ];
    }
  list_number( g_pa_last_ptr[ g_current_pa ] ) = 0;

  //Rs[
  list_number( index ) = line_number;
  list_length( index ) = new_line_size;
  dst = index + sizeof(int) + sizeof(char);
  new_line_buf += sizeof(int) + sizeof(char);
  for(;;)
    {
      char c = *new_line_buf++;
      g_pa_p[ g_current_pa ][ dst++ ] = c;
      if( c == '\0' ) break;
    }
  return true;
}



// s폜
void
delete_line( int line_number )
{
  int index, src, dst, len, n, prev;


  index = find_edit_line( line_number ); // sԍ

  if( list_number(index) == line_number )
    {
      dst = index;
      src = index + list_length( index );
      len = list_length( index );

      while( src < g_pa_last_ptr[ g_current_pa ] )
        {
          g_pa_p[ g_current_pa ][dst++] = g_pa_p[ g_current_pa ][src++];
        }
      g_pa_last_ptr[ g_current_pa ] -= len;
      list_number( g_pa_last_ptr[ g_current_pa ] ) = 0;
      g_pa_curr_number = 0; // ҏWƃZbg
      g_pa_num_lines[ g_current_pa ]--;

      // ŏIs
      if( line_number == g_pa_last_number[ g_current_pa ] )
        {
          index = prev = 0;
          for(;;)
            {
              n = list_number( index );
              if( n == 0 ) break;
              prev = index;
              index += list_length(index);
            }
          g_pa_last_number[ g_current_pa ] = list_number( prev );
        }
    }
}



// sԍ
// vOGAAŏ̍sԍ ȏ̍sTB
// num Tsԍ
// return s̈ʒu
// ȂAŌ̈ʒu
static int
find_edit_line( int num )
{
  int index = 0;
  int n;


  for(;;)
    {
      n = list_number( index );
      if( n == 0 || n >= num ) break;
      index += list_length(index);
    }

  if( num > g_pa_last_number[ g_current_pa ] )
    g_pa_last_number[ g_current_pa ] = num;

  return index;
}



// GOTOp s
static int
find_goto_line( int num )
{
  int index = 0;
  int n;


  for(;;)
    {
      n = list_number( index );
      if( n == num ) return index;
      else if( n == 0 ) break;
      index += list_length(index);
    }
  return -1;
}



// tB[hNA
void
flash_field( FieldPtr field_p )
{
  FldFreeMemory( field_p );
  FldDrawField( field_p );
}



// Ԍɕϊ
//
// obt@̍\:
//	int sԍ	1`9999
//	char		Xe[gg̃oCgB0`80
//	char * n	Xe[gg
//	char 		NULLI[ iACg܂j
void
conv_ascii_to_basic( void )
{
  FieldPtr field_p;
  char *read_p;
  char new_line_buf[ MAX_LINE_SIZE ];
  int new_line_size = 0;
  int line_number = 0;

  field_p = GetObjectPtr( MainDescriptionField );
  read_p = FldGetTextPtr ( field_p );
  if( ! read_p ) return;

  // sԍ
  line_number = get_line_number( &read_p );
  if( ! line_number )
    {
      // sԍȂII@G[
      disp_error( ERR_UL );
      return;
    }
  else if( line_number == -1 )
    return;

  *( (int *) new_line_buf ) = line_number;
  new_line_size += sizeof(int) + sizeof(char); //sԍ 2bytes + sTCY 1bytes

  // s܂ŒԌɕϊ
  if( ! conv_il( &read_p, new_line_buf, new_line_size ) )
    {
      disp_error( ERR_IL );
      return;
    }

  new_line_size = new_line_buf[2];
  // vOGAɑ}
  if( ! insert_pa( line_number, new_line_buf, new_line_size ) )
    {
      disp_error( ERR_OM );
      return;
    }

  // tB[hNA
  flash_field( field_p );
  // t[GATCY\
  update_free_area();
  // sZ
  g_pa_num_lines[ g_current_pa ]++;

  g_pa_curr_number = 0; // ݂̕ҏWƃZbg

  return;
}



// G[_CAO
static void
disp_error( int code )
{
  char s[5];
  char mes[][32] = {
    "Syntax error",
    "Out of memory",
    "GOSUB or FOR nesting exceeded",
    "RETURN without GOSUB",
    "Expression too complex",
    "Illegal line number",
    "Division by zero",
    "Index out of range",
    "Variable type mismatch",
    "String too long",
    "NEXT without FOR",
    "Data out of range",
    "Undefined line"
    };

  if( g_run_mode )
    {
      // G[s\
      g_err_number = *((int *)g_line_top);
      g_pa_curr_number = g_err_number;
    }

  FrmCustomAlert( basic_error, &mes[code][0], StrIToA( s, g_pa_curr_number ), NULL );
  if( g_run_mode )
    {
      g_run_mode = false;
      FrmGotoForm( MainForm );
    }
}



// vOGA
void
init_program_area( void )
{
  int i;


  g_run_mode = false; // ҏW[h RUN | EDIT
  g_current_pa = 0; // ݂̃vOGA p0
  g_pa_curr_ptr = 0;
  g_pa_curr_number = 0;

  for( i = 0; i < MAX_PROGRAM_AREA_NUM; i++ )
    {
      g_pa_free_size[i] = MAX_PROGRAM_SIZE;
      g_pa_last_ptr[i] = 0; // vOTCY
      g_pa_last_number[i] = 0;
      g_pa_num_lines[i] = 0; // s
    }
  //MemSet( g_pa_p, sizeof(char) * MAX_PROGRAM_AREA_NUM * MAX_PROGRAM_SIZE, 0 );
}


void
new_list( void )
{
  g_pa_curr_ptr = 0;
  g_pa_curr_number = 0;

  g_pa_free_size[ g_current_pa ] = MAX_PROGRAM_SIZE;
  g_pa_last_ptr[ g_current_pa ] = 0; // vOTCY
  g_pa_last_number[ g_current_pa ] = 0;
  g_pa_num_lines[ g_current_pa ] = 0; // s

  MemSet( &g_pa_p[ g_current_pa ][0], MAX_PROGRAM_SIZE, 0 );// NA
  update_free_area();
  flash_field( (FieldPtr) GetObjectPtr( MainDescriptionField ) );
}



// t[GA
void
update_free_area( void )
{
  char s[5];
  RectangleType r = { { 50,2 }, { 28, 10 } };


  g_pa_free_size[ g_current_pa ] = MAX_PROGRAM_SIZE - 2 - g_pa_last_ptr[ g_current_pa ];

  StrPrintF( s, "%04d", g_pa_free_size[ g_current_pa ] );
  WinEraseRectangle ( &r, 0 );
  WinDrawChars( s, StrLen( s ), 50, 2 );
}



// s\
void
disp_edit_line( void )
{
  int index = 0;
  int n;


  for(;;)
    {
      n = list_number( index );
      if( n == g_pa_curr_number ) break;
      index += list_length(index);
    }

  conv_basic_to_ascii( index );
}



// ЂƂO̍sҏW
void
backward_edit_line( void )
{
  int index = 0;
  int prev_index = 0;
  int n;
  FieldPtr field_p;


  if( ! g_pa_curr_number )
    g_pa_curr_number = g_pa_last_number[ g_current_pa ] + 1;

  for(;;)
    {
      n = list_number( index );
      if( n == 0 || n == g_pa_curr_number ) break;
      prev_index = index;
      index += list_length(index);
    }

  if( index != prev_index )
    {
      index = prev_index;
      g_pa_curr_number = list_number( index );
      conv_basic_to_ascii( index );
    }
  else
    {
      // VKtB[h
      g_pa_curr_number = 0; // ҏWsZbg
      field_p = GetObjectPtr( MainDescriptionField );
      FldFreeMemory( field_p );
      FldDrawField( field_p );
    }
}



// ЂƂ̍sҏW
void
forward_edit_line( void )
{
  int index = 0;
  int n;
  FieldPtr field_p;


  for(;;)
    {
      n = list_number( index );
      if( n == 0 || n > g_pa_curr_number ) break;
      index += list_length(index);
    }

  if( n )
    {
      g_pa_curr_number = n;
      conv_basic_to_ascii( index );
    }
  else
    {
      // VKtB[h
      g_pa_curr_number = 0; // ҏWsZbg
      field_p = GetObjectPtr( MainDescriptionField );
      FldFreeMemory( field_p );
      FldDrawField( field_p );
    }
}



// ԃR[heLXgɕϊ
// tB[hɏo
static void
conv_basic_to_ascii( int index )
{
  FieldPtr field_p = GetObjectPtr( MainDescriptionField );
  char c;
  int pos, i;
  char buf[128];


  FldFreeMemory( field_p );
//  FldDrawField( field_p ); // tB[hNA

  buf[0] = '\0';
  // sԍ
  StrIToA( buf, list_number(index) );
  StrCat( buf, " " );
  index += sizeof(int) + sizeof(char);

  for(;;)
    {
      c = g_pa_p[ g_current_pa ][ index++ ];
      if( c == '\0' ) break;

      for( i = 0; ; i++ )
        {
          if( (UInt8)c == keyword[i].code )
            {
              StrCat( buf, &keyword[i].word[0] );
              if( c != TOKEN_STRING && (UInt8)c != TOKEN_REM && (UInt8)c < TOKEN_ROPMIN || (UInt8)c > TOKEN_ROPMAX )
                StrCat( buf, " " );
              if( (UInt8)c == TOKEN_REM )
                {
                  StrCat( buf, &g_pa_p[ g_current_pa ][ index ] ); // s܂ŃRs[
                  goto skip;
                }
              else if( c == TOKEN_STRING )
                {
                  int start = index;
                  for(;;)
                    {
                      if( g_pa_p[ g_current_pa ][ index ] == '\"' || g_pa_p[ g_current_pa ][ index ] == '\0' ) break;
                      index++;
                    }
                  index++;
                  StrNCat( buf, &g_pa_p[ g_current_pa ][ start ], StrLen(buf)+(index-start+1) ); //  " ܂ŃRs[
                }
              break;
            }
          if( keyword[i].code == 0 )
            {
              pos = StrLen( buf );
              buf[ pos ] = c;
              buf[ pos + 1 ] = '\0';
              break;
            }
        }
    }
skip:
  FldInsert( field_p, buf, StrLen( buf ) );
  FldDrawField( field_p );
}



//////////////////////////////////////////////////////////////////////////////
// basic R}h

// 
static void
cmd_let( char **read_p )
{
  TokenType t = get_token( read_p );
  int v, code;

  if( g_err_number ) return;
  v = 0;
  switch( t.code )
    {
    case TOKEN_VAR:
    case TOKEN_VAR_S:
      v = t.value; // ӂ̕ϐ

      v += get_array_index( read_p ); // z̓Y
      if( g_err_number ) return;

      code = t.code;
      break;

    case TOKEN_VAR_D: // p
      code = t.code;
      break;

    default:
      disp_error( ERR_SE );
      return;
    }

  t = get_token( read_p );
  if( g_err_number ) return;

  if( t.code != TOKEN_EQ )
    {
      disp_error( ERR_SE );
      return;
    }

  if( code == TOKEN_VAR )
    cmd_let2( v, read_p );
  else if( code == TOKEN_VAR_S )
    cmd_let_str( v, read_p );
  else
    cmd_let_str_d( read_p );
}



static void
cmd_let2( int v, char **read_p )
{
  double value = expression( read_p );

  if( g_err_number ) return;
  if( v >= 0 && v < 26 )
    {
      if( g_variable[ v ].type == VARIABLE_STR ) // ^`FbN
        {
          disp_error( ERR_TM );
          return;
        }
      g_variable[ v ].contents.val = value;
      g_variable[ v ].type = VARIABLE_NUM; // ϐ̐ݒ
    }
  else
    {
      disp_error( ERR_OR );
      return;
    }

  check_end( *read_p );
}



static void
cmd_let_str( int v, char **read_p )
{
  char s[ MAX_STRING_VAR+1 ];


  str_expression( read_p, s ); // ̉ZiAj a$ = bS + c$ ܂ a$ = "aa" + b$ Ȃ

  if( g_err_number ) return;
  if( v >= 0 && v < 26 )
    {
      if( g_variable[ v ].type == VARIABLE_NUM ) // ^`FbN
        {
          disp_error( ERR_TM );
          return;
        }
      g_variable[ v ].contents.str[0] = '\0';
      StrCat( &g_variable[ v ].contents.str[0], s );
      g_variable[ v ].type = VARIABLE_STR; // ϐ̐ݒ
    }
  else
    {
      disp_error( ERR_OR );
      return;
    }

  check_end( *read_p );
}



// p
static void
cmd_let_str_d( char **read_p )
{
  char s[ MAX_STRING_VAR+1 ];


  str_expression( read_p, s );

  if( g_err_number ) return;

  g_special_strings[0] = '\0';
  StrCat( g_special_strings, s );

  check_end( *read_p );
}



// I
static void
cmd_end( char **read_p )
{
  static char s[] = "Done.";
  FieldPtr field_p;


  check_end( *read_p );
  if( g_err_number ) return;

  g_token_top = NULL;
  *read_p = NULL;

  // bZ[W
  field_p = GetObjectPtr( run_input );
  flash_field( field_p );
  FldInsert( field_p, s, StrLen(s) );
  SndPlaySystemSound( sndInfo );
}



// ϐNA
static void
cmd_vac( char **read_p )
{
  init_variable();
  g_cursor.x = g_cursor.y = 0; // J[\Zbg
  g_for_sp = g_go_sp = 0; // X^bNZbg
  MemSet( g_go_stack, sizeof(g_go_stack), 0 );
  MemSet( g_for_stack, sizeof(g_for_stack), 0 );

  check_end( *read_p );
}



// [P[g
static void
cmd_locate( char **read_p )
{
  if( get_token_char( read_p ) == ',' )
    {
      (*read_p)++;
      g_cursor.y = (Coord) expression( read_p );
      if( g_err_number ) return;
    }
  else
    {
      g_cursor.x = (Coord) expression( read_p );
      if( g_err_number ) return;
      if( get_token_char( read_p ) == ',' )
        {
          (*read_p)++;
          g_cursor.y = (Coord) expression( read_p );
          if( g_err_number ) return;
        }
    }
  check_end( *read_p );
}



// IF THEN THEN;őp
static void
cmd_if( char **read_p )
{
  TokenType t;
  Boolean b;


  t = get_token( read_p );
  if( g_err_number ) return;

  switch( t.code )
    {
    case TOKEN_KEY: // ֐
    case TOKEN_MID:
    case TOKEN_STR:
    case TOKEN_VAR_S: // ϐ
    case TOKEN_VAR_D: // pϐ$
    case TOKEN_STRING: // ""
      back_token( read_p );
      b = str_logic( read_p );
      if( g_err_number ) return;
      break;

    default:
      back_token( read_p );
      b = logic( read_p );
      if( g_err_number ) return;
      break;
    }

  if( b )
    {
      t = get_token( read_p );
      if( g_err_number ) return;

      if( t.code != TOKEN_THEN && t.code != TOKEN_SEMICOLON )
        {
          disp_error( ERR_SE );
          return;
        }
    }
  else
    {
      while( **read_p ) (*read_p)++; // s܂ŃXLbv
    }
}



// PRINT lϐ|ϐ|"" ; <JԂ>
static void
cmd_print( char **read_p )
{
  TokenType t;
  PointType save;
  double n;
  char s[ MAX_STRING_VAR+1 ];


  save.x = g_cursor.x; save.y = g_cursor.y;

  t = get_token( read_p );
  if( g_err_number ) return;
  if( t.code == TOKEN_NULL ) // PRINT ݂̂ cls
    {
      cmd_cls( read_p );
      return;
    }
  back_token( read_p );

  for(;;)
    {
      t = get_token( read_p );
      if( g_err_number ) return;
      if( t.code == TOKEN_NULL ) break;

      switch( t.code )
        {
        case TOKEN_KEY: // ֐
        case TOKEN_MID:
        case TOKEN_STR:
        case TOKEN_VAR_S: // ϐ
        case TOKEN_VAR_D: // pϐ$
        case TOKEN_STRING: // ""
          back_token( read_p );
          str_expression( read_p, s );
          if( g_err_number ) return;
          palm_put_str( s );
          break;

        case TOKEN_VAR: // lϐ
        case TOKEN_NUMBER: // l
        default:
          back_token( read_p );
          n = expression( read_p );
          if( g_err_number ) return;
          palm_put_num( n );
          break;
        }

      t = get_token( read_p );
      if( g_err_number ) return;
      if( t.code == TOKEN_NULL) break;

      switch( t.code )
        {
        case TOKEN_SEMICOLON:
          if( get_token_char( read_p ) == 0 ) return;
          break;

        default:
          disp_error( ERR_SE );
          return;
        }
    }

  g_cursor.x = save.x; g_cursor.y = save.y;
  check_end( *read_p );
}



// input (""|?),(ϐ|lϐ)
static void
cmd_input( char **read_p )
{
  FormPtr frm_p;
  UInt16 fieldIndex;
  TokenType t;
  char *save_p = (*read_p) -1;
  int v;
  PointType save;
  char * buf;
  char s[ MAX_STRING_VAR+1+2 ];


  save.x = g_cursor.x; save.y = g_cursor.y;
  (*read_p) = (*read_p) + g_input_skip; // p[^΍

  t = get_token( read_p );
  if( g_err_number ) return;

  if( t.code == TOKEN_STRING )
    {
      copy_str( read_p, s );
      palm_put_str( s );
      if( get_token_char( read_p ) == ',' )
        (*read_p)++;
      else
        {
          disp_error( ERR_SE );
          return;
        }
    }
  else
    {
      palm_put_str("?");
      back_token( read_p );
    }

  if( g_input_state == inp_off )
    {
      frm_p = FrmGetActiveForm();
      fieldIndex = FrmGetObjectIndex( frm_p, run_input );
      FrmSetFocus( frm_p, fieldIndex );

      SndPlaySystemSound( sndConfirmation ); // ͎tJn
      g_input_state = inp_on;
      g_input_buf[0] = '\0';
      g_cursor.x = save.x; g_cursor.y = save.y;
      *read_p = save_p;
      g_input_skip = 0;
      return;
    }
  else if( g_input_state == inp_end )
    {
      t = get_token( read_p );
      if( g_err_number ) return;

      switch( t.code )
        {
        case TOKEN_VAR_S: // ϐ
        case TOKEN_VAR: // lϐ
          v = t.value;
          v += get_array_index( read_p );
          if( g_err_number ) return;

          if( t.code == TOKEN_VAR )
            {
              buf = g_input_buf;
              if( StrLen( buf ) )
                cmd_let2( v, &buf );
            }
          else
            {
              s[0] = '\0';
              StrCat( s, "\"" );
              StrCat( s, g_input_buf ); // " "t
              StrCat( s, "\"" );
              buf = s;
              cmd_let_str( v, &buf );
            }

          if( g_err_number ) return;
          break;

        case TOKEN_VAR_D: // pϐ$
          g_special_strings[0] = '\0';
          StrCopy( g_special_strings, g_input_buf );
          break;

        default:
          disp_error( ERR_SE );
          return;
        }
      t = get_token( read_p );
      if( g_err_number ) return;

      switch( t.code )
        {
        case TOKEN_SEMICOLON:
          g_input_state = inp_off;
          check_end( *read_p );
          input_end();
          return;

        case TOKEN_COMMA:
          g_input_state = inp_continue;
          g_input_skip = (UInt32)*read_p - (UInt32)save_p -1;
          g_cursor.x = save.x; g_cursor.y = save.y;
          *read_p = save_p;
          return;

        case TOKEN_NULL:
          g_input_state = inp_off;
          g_cursor.x = save.x; g_cursor.y = save.y;
          check_end( *read_p );
          input_end();
          return;

        default:
          disp_error( ERR_SE );
          return;
        }
    }

  *read_p = save_p;
  g_cursor.x = save.x; g_cursor.y = save.y;
  return;
}



static void
input_end( void )
{
  FieldPtr field_p;


  field_p = GetObjectPtr( run_input );
  FldReleaseFocus( field_p );
}



// o
static void
palm_put_str( char *s )
{
  WinDrawChars( s, StrLen(s), g_cursor.x, g_cursor.y );

  g_cursor.x += FntCharsWidth( s,StrLen(s) );
}



// lo ȊO  \
static void
palm_put_num( double d )
{
  FlpCompDouble f;
  char s[14+1];


  if(d - (int) d != 0 )
    {
      f.d = d;
      FlpFToA( f.fd, s );
    }
  else
    {
      StrIToA( s, (int)d );
      StrCat( s, "." );
    }

  WinDrawChars( s, StrLen(s), g_cursor.x, g_cursor.y );

  g_cursor.x += FntCharsWidth( s,StrLen(s) );
}



// cls
static void
cmd_cls( char **read_p )
{
  RectangleType r;

  r.topLeft.x = 0; r.topLeft.y = 0; r.extent.x = 159; r.extent.y = 159-12;
  WinEraseRectangle( &r, 0 );

  check_end( *read_p );
}



// goto 
void cmd_goto( char **read_p )
{
  int n;


  n = (int) expression( read_p );
  if( g_err_number ) return;

  n = find_goto_line( n );

  if( n < 0 )
    {
      *read_p = NULL;
      disp_error( ERR_IL );
      return;
    }

  check_end( *read_p );
  if( g_err_number ) return;

  g_line_top = &g_pa_p[ g_current_pa ][ n ];
  *read_p = g_line_top + sizeof(int) + sizeof(char);
}



// gosub 
static void
cmd_gosub( char **read_p )
{
  int n;


  if( g_go_sp >= GO_STACK_SIZE )
    {
      disp_error( ERR_MG );
      return;
    }

  n = (int) expression( read_p );
  if( g_err_number ) return;

  n = find_goto_line( n );
  if( n < 0 )
    {
      *read_p = NULL;
      disp_error( ERR_IL );
      return;
    }

  check_end( *read_p );
  if( g_err_number ) return;

  g_go_stack[g_go_sp][0] = g_line_top;
  g_go_stack[g_go_sp][1] = *read_p;
  g_go_sp++;

  g_line_top = &g_pa_p[ g_current_pa ][ n ];
  *read_p = g_line_top + sizeof(int) + sizeof(char);
}



// return
static void
cmd_return( char **read_p )
{
  if( g_go_sp == 0 )
    {
      disp_error( ERR_WG );
      return;
    }

  check_end( *read_p );
  if( g_err_number ) return;

  g_go_sp--;
  g_line_top = g_go_stack[ g_go_sp ][0];
  *read_p = g_go_stack[ g_go_sp ][1];
}



// for  to  step 
static void
cmd_for( char **read_p )
{
  TokenType t;
  int v;


  if( g_for_sp >= FOR_STACK_SIZE )
    {
      disp_error( ERR_MG );
    }

  t = get_token( read_p );
  if( g_err_number ) return;

  switch( t.code )
    {
    case TOKEN_VAR:
      v = t.value;
      v += get_array_index( read_p );
      if( g_err_number ) return;

      t = get_token( read_p );
      if( g_err_number ) return;
      if( t.code != TOKEN_EQ )
        {
          disp_error( ERR_SE );
          return;
        }
      cmd_let2( v, read_p );
      break;

    default:
      disp_error( ERR_SE );
      return;
    }

  // X^bNɐς
  g_for_stack[g_for_sp].top = g_line_top; // s
  g_for_stack[g_for_sp].v = v; // ϐ

  check_end( *read_p );
}


// to 
static void
cmd_to( char **read_p )
{
  TokenType t;
  double g, step;


  g = expression( read_p );
  if( g_err_number ) return;

  t = get_token( read_p );
  if( t.code == TOKEN_STEP )
    {
      step = expression( read_p );
    }
  else
    {
      back_token( read_p );
      step = 1.0;
    }

  // X^bNɐς
  g_for_stack[g_for_sp].g = g; // Il
  g_for_stack[g_for_sp].step = step; // ϐ
  g_for_stack[g_for_sp].read = *read_p; // ϐ
  g_for_sp++;

  check_end( *read_p );
}


// next ϐ
static void
cmd_next( char **read_p )
{
  TokenType t;
  double val, step;


  if( g_for_sp == 0 )
    {
      disp_error( ERR_NX );
      return;
    }

  t = get_token( read_p );
  if( t.code != TOKEN_VAR )
    {
      disp_error( ERR_SE );
      return;
    }
  if( t.value != g_for_stack[ g_for_sp-1 ].v )
    {
      disp_error( ERR_NX );
      return;
    }

  val = g_variable[ g_for_stack[g_for_sp-1].v ].contents.val;
  step = g_for_stack[g_for_sp-1].step;

  if( val == g_for_stack[g_for_sp-1].g )
    {
      g_for_sp--; // forI

      val += step;
      if( val - (int)val != 0 )
        val = round_n( val, ROUND_POW );
      g_variable[ g_for_stack[g_for_sp].v ].contents.val = val;
      check_end( *read_p );
      return;
    }
  val += step;
  if( val - (int)val != 0 )// 
    val = round_n( val, ROUND_POW );
  g_variable[ g_for_stack[g_for_sp-1].v ].contents.val = val;

  g_line_top = g_for_stack[ g_for_sp-1 ].top;
  *read_p = g_for_stack[ g_for_sp-1 ].read;
}


// COLOR
static void
cmd_color( char **read_p )
{
  int col;


  col = expression( read_p );
  if( g_err_number ) return;

  WinSetTextColor( (IndexedColorType)col );

  check_end( *read_p );
}


// L[
static UInt8
cmd_key( void )
{
  UInt16 key;


  key = g_key_code;
  g_key_code = 0;

  switch( key )
    {
    case vchrPageUp:
      key = '\xb4';//
      break;
    case vchrPageDown:
      key ='\xb2';//
      break;
    case vchrHard1:
      key ='\xa1';//
      break;
    case vchrHard2:
      key ='\xa2';//
      break;
    case vchrHard3:
      key ='\xa3';//
      break;
    case vchrHard4:
      key ='\xa4';//~
      break;
    }

  return (UInt8)key;
}



// ֐
// 
static int
cmd_len( char **read_p )
{
  char s[ MAX_STRING_VAR+1 ];

  return str_expression( read_p, s );
}


// 𐔒lɕϊ
static double
cmd_val( char **read_p )
{
  char s[ MAX_STRING_VAR+1 ];
  FlpCompDouble d;


  str_expression( read_p, s );
  if( g_err_number ) return 0;

  d.fd =  FlpAToF( s );
  return d.d;
}


// l𐔎ɕϊ
static int
cmd_str( char **read_p, char *s )
{
  double v;
  FlpCompDouble f;


  v = expression( read_p );
  if( g_err_number )
    {
      s[0] = '\0';
      return 0;
    }

  // or
  if(v - (int) v != 0 )
    {
      f.d = v;
      FlpFToA( f.fd, s );
    }
  else
    {
      StrIToA( s, (int)v );
    }

  return StrLen( s );
}



//wʒu̕o 1`MAX_STRING_VAR+1
static int
cmd_mid( char **read_p, char *s )
{
  int v,n;
  TokenType t;


  v = expression( read_p );// ʒuij
  if( g_err_number ) return 0;

  t = get_token( read_p );
  if( g_err_number ) return 0;
  if( t.code != TOKEN_COMMA )
    {
      back_token( read_p );
      n = MAX_STRING_VAR;
    }
  else
    {
      n = expression( read_p );
      if( g_err_number ) return 0;
    }

  if( !v ) return 0;
  if( --v > MAX_STRING_VAR ) return 0;
  if( v > StrLen( g_special_strings ) ) return 0;

  // Rs[
  StrNCat( s, g_special_strings + v, n+1 );

  return StrLen( s );
}


// s`FbN
static void
check_end( char *read_p )
{
  char c;


  c = *read_p;
  if( c != '\0' && (UInt8)c != TOKEN_REM && (UInt8)c != TOKEN_COLON && (UInt8) c != TOKEN_TO )
    {
      disp_error( ERR_SE );
      return;
    }
}

//  Rs[
static int
copy_str( char **read_p, char *s )
{
  int len = 0;
  char *src = (*read_p);

  s[0] = '\0';
  for(;;)
    {
      if( **read_p == '\0' )
        {
          s[0] = '\0';
          disp_error( ERR_SE );
          return 0;
        }
      if( **read_p == '\"' ) break;
      len++;
      (*read_p)++;
    }
  (*read_p)++;

  if( len > MAX_STRING_VAR )
    {
      len = MAX_STRING_VAR;
      disp_error( ERR_LG );
      return 0;
    }

  StrNCat( s, src, len+1 );
  return len;
}


#if 0
// A TO F
double
a_to_f( char *s )
{
  double val, power;


  val = power = 0.0;
  if( *s < '0' || *s > '9' || *s == '\0' ) return 0;

    for(;;)
    {
      val = val * 10.0 + ( *s - '0');
      power *= 10.0;
      if( power > 10000000000.0 || val > DBL_MAX ) // double 10܂ŔF
        {
          power = 10000000000.0;
          break;
        }
      s++;
      if( *s == '.' )
        {
          power = 1.0;
          s++;
        }
      if( *s < '0' || *s > '9' || s == '\0' ) break;
    }
  if( power )
    val /= power;


  return val;
}

#endif
